// *************************************************************************
//    ExecutionEnv.h
// *************************************************************************

#ifndef _ExecutionEnv_H_
#define _ExecutionEnv_H_

#include "RelationMisc.h"
#include "StackMark.h"
#include "CoralStream.h"

// *************************************************************************

enum ExecutionMode {
    COR_DOING_MAGIC,
    COR_DOING_INTERPRET,
    COR_DOING_COMPILE,
    COR_DOING_FACTOR,
    COR_DOING_FACTOR_MAGIC,
    COR_DOING_IMPERATIVE,
};

enum RelationTypeDefault {
    COR_LIST_RELATION,
    COR_HASH_RELATION,
};

// *************************************************************************
// The execution environment contains all the globals that represent
// the state of CORAL execution. This includes default values, as
// well as debugging and profiling flags. It should probably also
// contain the Trail, and the StorageManager. The constructor should
// create a StackMark, so that if for some reason (like a signal)
// CORAL needs to be restarted, it can be done by popping the
// StackMark ---- PRAVEEN

class Cor_ExecutionEnv {

  public:
    ExecutionMode C_exec_mode;
    RelationOptions GlobalRelationOptions;
    char *coral_path;
    int SM_initialized;

    // used to keep the bindenv stack consistent in event of failures
    Cor_StackMark *stackmark;

    int C_insert_mode_default;
    int C_quiet_mode_default;
    int C_answer_style_default;
    int C_use_quotes;
    int C_read_coralrc_default;
    int C_create_magic_default;
    int C_verbose_answer_default;
    int C_verbose_aggsel_default;

    int C_print_prompt_default;
    int C_do_history_default;

    int C_rewriting_default;
    int C_use_supplementary_magic_default;
    int C_use_magic_default;
    int C_use_factor_magic_default;
    int C_use_factoring_default;
    int C_use_pipelining_default;
    int C_use_ordsearch_default;
    int C_use_lazy_eval_default;
    int C_use_monotonic_default;
    int C_interactive_mode_default;
    int C_interactive_mode_current;
    int C_check_subsum_default;
    int C_multisets_default;
    int C_index_deltas_default;
    int C_use_return_unify_default;
    int C_diskrel_default;
    int C_eager_indexing_default;
    int C_sip_reorder_default;

    int C_preprocessing_default;
    int C_single_scc_default;
    int C_convert_functions_default;
    int C_single_answer_default;
    int C_naive_bcktrking_default;
    int C_nested_scc_eval_default;

    // variables used for debugging
    int dbg_unify ;
    int dbg_subsumes;
    int dbg_indexing;
    int dbg_aggregates;
    int dbg_get_next;
    int dbg_sets;
    int dbg_ordsearch;
    int dbg_return_unify;
    int dbg_pipelining;
    int dbg_backtrack;
    int dbg_deletes;
    int dbg_modules;

    int C_hash_buckets_default;    
    int C_delta_buckets_default;
    double C_incr_ratio_default;
    double C_max_occupancy_default;
    int C_max_recursion_depth_default;
    RelationTypeDefault C_relation_type_default;
    int C_connect_exodus_default;

      // variables used for profiling
    int profile_sn;
    int profile_scc;
    int profile_module;

    CoralStream *in_strm;
    CoralStream *out_strm;
    CoralStream *err_strm;

    CoralStream *explain_strm;
    CoralStream *trace_strm;
    CoralStream *profile_strm;

    char *in_filename;
    char *out_filename;
    char *error_filename;
    char *trace_filename;
    char *profile_filename;

    int C_interrupt_raised;
    // Commands to be processed when back in yyparse() 
    int quit;              // Should we quit
    int yield;             // Should we yield after this command

    // A few parse items that cannot be in the ParserEnv
    int C_at_end_module;
    int in_query_loop;
    int c_module_separator;
    
    void init(CoralStream *in=NULL, CoralStream *out=NULL,
            CoralStream *err=NULL);
    void display(CoralStream *outf);
    Cor_ExecutionEnv(Cor_ExecutionEnv *proto,
            CoralStream *in, CoralStream *out, CoralStream *err);
    Cor_ExecutionEnv();
   ~Cor_ExecutionEnv();
    void setMode(ExecutionMode ex_mode) { C_exec_mode = ex_mode; }
    void flush_all_outputs();

      // For quickly testing whether we are running as magic
    int i_am_magic() {
        if ((C_exec_mode == COR_DOING_MAGIC)
                || (C_exec_mode == COR_DOING_FACTOR)
                || (C_exec_mode == COR_DOING_FACTOR_MAGIC))
            return(1);
        return(0);
    }

    char *on_off_str(int flag) {
        return( flag ? "yes" : "no" );
    }
};

// *************************************************************************

#define COR_BINDING_ANSWER_PRINT    0
#define COR_TUPLE_ANSWER_PRINT      1
#define COR_EMPTY_ANSWER_PRINT      2
#define COR_LITERAL_ANSWER_PRINT    3

// *************************************************************************

#endif _ExecutionEnv_H_
