// *************************************************************************
//    HashSimpleRelation.h
// *************************************************************************

#ifndef _HashSimpleRelation_H_
#define _HashSimpleRelation_H_

#include "StorageRelation.h"
#include "RelationList.h"
#include "HashIndex.h"
#include "HashIndexSet.h"
#include "Debug.h"

// *************************************************************************
//
//  Tuples which are newer than the oldest RMark are in the "chain" list
//  inherited from LinkedRelation.  The others are accessed through the
//  Hash Indexes.
//
// *************************************************************************

class Cor_HashSimpleRelation : public Cor_StorageRelation {

  public:

      // The index
    Cor_HashIndexSet index_set;


    Cor_HashSimpleRelation( int arity, int delta = -1, int buckets = -1);
    virtual ~Cor_HashSimpleRelation();

    void initialize_iterator(Cor_TupleIterator&);

    virtual void freeMark(Cor_RMark *);
    virtual void free_allMarks();

      // Indexing related methods
    virtual Cor_GenericIndexSet *get_index_set() { return &index_set; }
    virtual Cor_GenericIndex *create_index_object(Cor_BitVector *bv) {
        return new Cor_HashIndex(this, bv, -1);
    }

    virtual Cor_GenericIndex *create_index_object(Cor_ArgList *pat, int n_var,
        Cor_ArgList *v_nm) {
        return new Cor_HashIndex(this, pat, n_var, v_nm);
    }


    virtual int insert_tuple(Cor_Tuple *tuple);
    virtual int insert_raw( Cor_Tuple *tuple );

    virtual Cor_Tuple *get_next_tuple( Cor_TupleIterator& it);
    virtual void release(Cor_TupleIterator& it);
    virtual Cor_BindEnv *get_next( Cor_TupleIterator& );
    virtual int is_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark = NULL,
        Cor_RMark *endmark = NULL);

    virtual void free_all();
    virtual void free_indices();
    virtual void free_delta_relations();

      // Move all tuples/index entries into rel.  
      // Assumes that rel has the same set of indices, 
      // and all corresponding indices and relations are
      // of exactly the same type.
    int move_all_entries(Cor_StorageRelation *rel);     

    virtual void empty_relation(int deleteTuples=0);
    virtual void print_name(ostream *out);

    virtual void show_internals(ostream& out);

#ifdef ASYN_SOCK_REL
    virtual void get_tuples(void);
#endif
};

// *************************************************************************

#endif _HashSimpleRelation_H_
