// *************************************************************************
//    SccInfo.h
// *************************************************************************

#ifndef _SccInfo_H_
#define _SccInfo_H_

//#include "Table.h"
#include "Predicate.h"
#include "QFMaterializedModuleData.h"
#include "Scc.h"
#include "SemiNaive.h"

#ifdef ASYN_SOCK_REL
#include "PredicateList.h"
#endif

// *************************************************************************
//    Codependent classes

class Cor_SemiNaive;
class Cor_QFMaterializedModuleData;
class Cor_Table;

// *************************************************************************

class Cor_SccInfo {

  public:

    Cor_QFMaterializedModuleData *module;  // Module to which it belongs.
    Cor_Scc *scc;
    Cor_SemiNaive *sn_rules;           // Array of seminaive rules
    int numSNRules;
    Cor_Predicate **pred_ptrs;         // Array of preds corresponding to
                                       // the scc
    int numPreds;                      // number of preds in this SCC
    int num_lowerPreds;                // number of preds in all previous SCCs
    int *offset_arr;                   // Offsets of preds in RMarkArray
                                       // Size of this is numPreds
    int *sn_offset_arr;                // Positions of heads of sn rules
                                       // in pred_ptrs
    int *pred_success_arr;             // Number of newtuples generated for
                                       // each predicate in the previous
                                       // iteration size is : numPreds
    int rec_rule_start;                // Starting position of recursive rules
    int scc_id;                        // number identifying the position of
                                       // this scc in the sccArray of the
                                       // module

#ifdef ASYN_SOCK_REL
    int numSockRelPreds;               // number of preds of socket_rel

    int              external_flag_on();
    int              numPendingSockRel();

    Cor_PredicateList *SockRelPreds;
#endif
    
    Cor_SccInfo(Cor_QFMaterializedModuleData *md, Cor_Scc *scc,
        Cor_Predicate *, int id);
   ~Cor_SccInfo();

        // allocates memory for seminaive rules, and returns the number
        // of rules in the scc. This method also sets the value of numSNRules
    int  allocateRuleStructures(Cor_Predicate *);

    void Solve(Cor_Table *F, Cor_TupleIterator *iterator = NULL);
    void clean_cache() {
        if (sn_rules) sn_rules->clean_cache();
    }

    int iterateRecursiveRules(Cor_Table *F, Cor_RMarkArray *rm_Arr, int);
    int slide_deltas(Cor_Table *F, Cor_RMarkArray *rm_Arr, int free_all);

    void print(ostream *fd, int print_preds = 1, 
         int print_text = 1, int print_snrules = 1);

  private:
    void createSNStructs(Cor_Predicate *);
    Cor_RMarkArray *setupSolve(Cor_Table *F);
    void display_annotation_info(Cor_Table *F);
};

// *************************************************************************

#endif _SccInfo_H_
