// *************************************************************************
//  Entity.h 
// *************************************************************************

#ifndef ENTITY_H
#define ENTITY_H

#include <stdlib.h>
#include "Debug.h"

//#define    SHOW_ALLOC    1

class  Cor_Entity;
class  Cor_Entity_Array;
extern Cor_Entity_Array *Coral_Free;

        // Allocate everything in chunks that are multiples of
        // ALIGN_SZE, so all chunks in a slot are homogenous
        // This should be at least 8, and be a power of two 
        // (see sz_ceil() method)
#define    ALIGN_SZE        (8)
//#define    ALIGN_SZE        (sizeof(double))

        // This is the number of slots you want - adjustable
#define    ENTITY_ARRY_SZE             (256)


        // This is the maximum chunk size that will have a slot available
#define    MAX_CHUNK_SZE     (ALIGN_SZE * ENTITY_ARRY_SZE)


    // A bogus structure to save space in the Entity class
typedef struct Ent_Element {
    struct Ent_Element    *next;    // A pointer to the next free block
} Cor_Ent_Element;


class Cor_Entity_Array {

  public:
                // A table of chunks.
                // Chunks are stored based on size in multiples
                // of ALIGN_SZE:
                //    chnk_tab[i] : List of chunks of size ((i+1)*ALIGN_SZE) 
    Cor_Ent_Element  *chnk_tab[ENTITY_ARRY_SZE];

    int           num_alloc[ENTITY_ARRY_SZE];  // The total number allocated
    int           num_free[ENTITY_ARRY_SZE];   // and present in free list


  private:
      // Round up all sizes to a multiple of ALIGN_SZE
    size_t sz_ceil(size_t size) {
        // A faster way?
        return( (size + ALIGN_SZE-1) & ~(ALIGN_SZE-1) );
    }

      // The incoming size should be a multiple of ALIGN_SZE
    int index(size_t size) {
        return((size / ALIGN_SZE) - 1);
    }


  public:
    Cor_Entity_Array() {
        int j;
        for (j=0; j < ENTITY_ARRY_SZE; j++) {
            chnk_tab[j] = NULL;
            num_alloc[j] = 0;
            num_free[j] = 0;
        }
    }

   ~Cor_Entity_Array() {
        delete_all();
    }

    void delete_all();

    void *acquire(size_t size);

    int  tmp_is_free(Cor_Ent_Element*);

    void dispose(Cor_Ent_Element *trash, size_t size);

      // For debugging
    void show_all();
};

/*
    // Use these to override a class's new/delete:
#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif
*/

// *************************************************************************

#endif  ENTITY_H
