// *************************************************************************
//  HashVal.h - Definitions for the hashing of arguments
// *************************************************************************

// *************************************************************************
// Hashing and HashConsing constants:
//    Any hash value other than VarHashValue and UnknownHashValue 
//        is a hash value of a ground term
//    A term whose hash value is VarHashValue is non-ground.
//    A term whose hash value is UnknownHashValue may be ground or non-ground
//            -- its hash value has not been computed.
//
// *************************************************************************

#ifndef    _HASH_VAL_H
#define    _HASH_VAL_H

typedef unsigned int HashVal;

#define VarHashValue ( (HashVal) -1)

  // This value is not generally supported!
#define UnknownHashValue ( (HashVal) -2)


HashVal Cor_StrHash(const char *name);

// *************************************************************************

inline int Cor_isGroundHashValue( HashVal i)
{
    return ((i != VarHashValue) && (i != UnknownHashValue));
}

// *************************************************************************
inline HashVal Cor_IntToHash(int i)
{
    HashVal hash = (HashVal) i;

     // This is done to insure that code that assumes hash functions
     // do not return UnknownHashValue or VarHashValue do not need
     // to be rewritten.

    if (hash >= UnknownHashValue) {
        hash = 1;
    }

    return(hash);
}

// *************************************************************************

#endif     _HASH_VAL_H
