// *************************************************************************
//    Collector.h
// *************************************************************************

#ifndef _Collector_H_
#define _Collector_H_

#include "HashSimpleRelation.h"
#include "QueryStack.h"
#include "User.h"
#include "Debug.h"

class Cor_BindEnv;

#define    INSERTS_BEFORE_YIELD    100

// *************************************************************************
//
// A Collector keeps track of the server-side of a single relation that
// is being solved by distributed evaluation with multiple Corals. (a reef)
//
// *************************************************************************

    // For simple producer/consumer style parallelism
enum Cor_CU_Type {
    Cor_Producer = 0,
    Cor_Consumer
};

class Cor_CollectUser {
  public:
    Cor_User        *user;
    Cor_CU_Type      type;

    Cor_QueryStack   queries;

    Cor_CollectUser(Cor_User *_user=NULL, Cor_CU_Type cutype=Cor_Producer);
};

class Cor_Collector {

  private:

    Cor_Relation      *datarel;         // The relation holding the data
                                        // for this collection.

    Cor_BindEnv       *qenv;            // A bindenv to use

    int num_users;      // The number of users in the array
    int length;         // Length of the array below

    int num_producers;    // The number of users of each type
    int num_consumers;    

    Cor_CollectUser  **userArray;       // An array of users involved
                                        // with this collection.

    int Find_User(Cor_User *user);
    void AddQuery(Cor_User *user, Cor_Tuple *query);
    void AddedFact(Cor_User *user, Cor_Tuple *fact);

  public:

    Cor_Collector(Cor_Relation *rel=NULL);
   ~Cor_Collector();

    Cor_Relation *rel() { return(datarel); }

    int  AddUser(Cor_User *user, Cor_CU_Type cutype);
    void DeleteUser(Cor_User *user);
    int  ProcessUser(Cor_User *user);
};

// *************************************************************************

#endif _Collector_H_
