// *************************************************************************
//    Annotations.C
// *************************************************************************

#include <stdlib.h>

#include "Annotations.h"

#include "PredAnnotations.h"
#include "StackMark.h"
#include "BindEnv.h"
#include "Globals.h"
#include "ParserStruct.h"
#include "Compile.h"
#include "GenericIndex.h"
#include "Unparse.h"
#include "Unify.h"

#include "Interface.h"
#include "Util.h"
#include "CArg.h"
#include "Debug.h"

// *************************************************************************
Cor_BitVector *Cor_create_adornment(char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int            j, len;
    Cor_BitVector *avect = NULL;

    len = strlen(adorn);

    if (len) {
        char *ap;
        avect = new Cor_BitVector(len);

        for(j=0, ap = adorn; j < len; j++, ap++) {
            if (*ap == 'b')
                avect->set(j);
        }
    }

    return(avect); 
}

// *************************************************************************
Cor_BitVector *Cor_create_adornment(Cor_CArg& adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! adorn.is_empty())
        return(Cor_create_adornment(adorn.symbol()));
    else
        return(NULL);
}

// *************************************************************************
int Cor_duplicate_index(Cor_PredAnnotations *annos, Cor_CArg& pred,
    Cor_BitVector *bv, Cor_ArgList *al1, int num_vars, Cor_ArgList *al2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stackmark;
    char buf[500];

    for (; annos; annos = annos->next()) {
        sprintf(buf, "%s", annos->pred.symbol());
        if (annos->adorn)
            sprint_adornment(buf+strlen(buf), annos->adorn);

        if (((pred.equals(annos->pred)) || (pred.equals(buf))) &&
            ((bv == annos->adorn) ||
            (bv && annos->adorn && (*bv == *(annos->adorn)))))
        { 
            if (!al2 && !annos->arglist2) {
                if ((*al1)[0] == ((*(annos->arglist1))[0]))
                    return 1;
            } else if (al2 && annos->arglist2) {
                Cor_BindEnv pat_env1(num_vars);
                Cor_BindEnv pat_env2(num_vars);
                Cor_BindEnv pat_env3(num_vars);
                Cor_BindEnv pat_env4(num_vars);

                // check if pattern and the var_names of each subsume 
                // that of the other

                if ((Cor_subsumes_arg_lists(*concat_arglists(al1, al2),
                    *concat_arglists(annos->arglist1, annos->arglist2),
                    &pat_env1, &pat_env2) == COR_U_SUCCEED) &&
                    (Cor_subsumes_arg_lists(
                        *concat_arglists(annos->arglist1, annos->arglist2),
                         *concat_arglists(al1, al2),
                        &pat_env3, &pat_env4) == COR_U_SUCCEED))
                {
                    stackmark.pop_to();
                    return 1;
                }
                stackmark.pop_to();
            }
        }
    }

    // for good measure
    stackmark.pop_to();
    return 0;
}

// *************************************************************************
/*
    WARNING: The bitvector passed to this function must be on static storage.
    It is not copied out.
    return value is whether the ArgLists were used.
*/
int  Cor_add_make_index_annotation(Cor_PredAnnotations *&annos, Cor_CArg *cp,
            Cor_CArg *ca, Cor_ArgList *a1, Cor_ArgList *a2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg p, a;

    if (cp) 
        p = *cp;
    if (ca) 
        a = *ca;

        // another function!
    return(Cor_add_make_index_annotation(annos, p, a, a1, a2));
};

int  Cor_add_make_index_annotation(Cor_PredAnnotations *&annos, Cor_CArg& pred,
                   Cor_CArg& adorn_string, Cor_ArgList *al1, Cor_ArgList *al2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_BitVector *bitv = Cor_create_adornment(adorn_string);

    char *argindex = NULL;
    int num_vars = 0;

    if (al2) {
        int mx1, mx2;
        mx1 = al1->max_vars();
        mx2 = al2->max_vars();
        num_vars = (mx1 > mx2) ? mx1 : mx2;
        argindex = Cor_simplify_index(al1, num_vars , al2);
    }

    if (argindex) {
        al1 = new Cor_ArgList(1);
        (*al1)[0] = Cor_CArg(argindex);
        al2 = NULL;
        delete [] argindex;
    } else {
        al1 = al1->dup();
        al2 = al2->dup();
    }

    if (! Cor_duplicate_index(annos, pred, bitv, al1, num_vars, al2)) {
    
        Cor_PredAnnotations *pa = new Cor_PredAnnotations(pred, annos);
            
        pa->arglist1 = al1;
        pa->arglist2 = al2;
        pa->adorn    = bitv;

        annos = pa;
    } else {
        delete al1;
        delete al2;
    }

    return(0);
}

// *************************************************************************
int  Cor_add_make_index_annotation(Cor_PredAnnotations*& annos, Cor_CArg& pred,
    Cor_BitVector *bv, Cor_ArgList *al1, Cor_ArgList *al2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    char *argindex = NULL;
    int num_vars = 0;

    if (al2) {
        int mx1, mx2;
        mx1 = al1->max_vars();
        mx2 = al2->max_vars();
        num_vars = (mx1 > mx2) ? mx1 : mx2;
        argindex = Cor_simplify_index(al1, num_vars, al2);
    }

    if (argindex) {
        al1 = new Cor_ArgList(1);
        (*al1)[0] = Cor_CArg(argindex);
        al2 = NULL;
        delete [] argindex;
    } else {
        al1 = al1->dup();
        al2 = al2->dup();
    }

    if (! Cor_duplicate_index(annos, pred, bv, al1, num_vars, al2)) {

            // LEAK
        Cor_PredAnnotations *pa = 
            new Cor_PredAnnotations (pred, annos);

        pa->arglist1 = al1;
        pa->arglist2 = al2;
        pa->adorn    = bv->dup();

        annos = pa;

    } else {
        delete al1;
        delete al2;
    }

    return(0);
}

// *************************************************************************
void Cor_add_agg_sel_annotation(Cor_ParserStruct *Cor_parserStruct,
        Cor_CArg *cPredName, Cor_CArg *cAdornStr,
        Cor_ArgList *al1, Cor_ArgList *al2, Cor_CArg *cOp_arg,
        AggSelType agg_type)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred;

    if (cPredName) 
      pred = *cPredName;

    Cor_PredAnnotations *pa =
        new Cor_PredAnnotations(pred, Cor_parserStruct->agg_sel_annotations);

    pa->arglist1 = al1;
    pa->arglist2 = al2;

    if (cAdornStr)
        pa->adorn  = Cor_create_adornment(*cAdornStr);
    else
        pa->adorn  = NULL;

    pa->op_arg = *cOp_arg;
    pa->agg_type = agg_type;

    Cor_parserStruct->agg_sel_annotations = pa;
}

// *************************************************************************
void Cor_add_allowed_adornment(Cor_ParserStruct *Cor_parserStruct, Cor_CArg *c,
    Cor_ArgList *args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_CArg pred_name;
    if (c)
        pred_name = *c;

    for (int i=0;i < args->count();i++) {
        if (! (*args)[i].is_symbol()) {
            *(Cor_exEnv->err_strm)
                << "Bad kind of adornment for allowed adorns "
                << pred_name.symbol() << ".\n";
            continue;
        }
        Cor_CArg adornname = (*args)[i];

        Cor_add_allowed_adornment(Cor_parserStruct, pred_name, 
                  adornname.symbol(),
                  adornname.symlen());
    }
}

// *************************************************************************
void Cor_add_allowed_adornment(Cor_ParserStruct *Cor_parserStruct,
        Cor_CArg& pred_name, char *adorn_string, int adorn_length)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *temp_adorn_string = adorn_string;

    for (int i=0; i < adorn_length; i++) {
        char adorn_char = *temp_adorn_string++;
        switch (adorn_char) {
        case 'b':
            break;
        case 'f':
            break;
        case 'a':
        {
            adorn_string[i] = 'b';
            Cor_add_allowed_adornment(Cor_parserStruct, pred_name, adorn_string,
                adorn_length);
            adorn_string[i] = 'f';
            Cor_add_allowed_adornment(Cor_parserStruct, pred_name, adorn_string,
                adorn_length);
              // Restore the string to its original state
            adorn_string[i] = 'a';
            return;
        }

        default:
            *(Cor_exEnv->err_strm)
                << "Bad character in adornment string '"
                << adorn_string << "'\n";
            return;
        }
    }

    // should get here only if all the characters in adorn_string are
    // either 'b' or 'f'

    Cor_PredAnnotations *pa = new Cor_PredAnnotations(pred_name,
        Cor_parserStruct->AllowedAdornList);

    pa->arglist1 = pa->arglist2 = NULL;
    pa->adorn = Cor_create_adornment(adorn_string);

    Cor_parserStruct->AllowedAdornList = pa;
}

// *************************************************************************
void Cor_add_disallowed_adornment(Cor_ParserStruct *Cor_parserStruct,
    Cor_CArg& pred_name, Cor_ArgList *args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (int i=0; i < args->count(); i++) {
        if (! (*args)[i].is_symbol()) {
            *(Cor_exEnv->err_strm)
                << "Bad kind of adornment for disallowed adorns "
                << pred_name.symbol() << ".\n";
            continue;
        }
        Cor_CArg adornname = (*args)[i];

        Cor_add_disallowed_adornment(Cor_parserStruct, pred_name,
                adornname.symbol(), adornname.symlen());
    }
}

// *************************************************************************
void Cor_add_disallowed_adornment(Cor_ParserStruct *Cor_parserStruct,
    Cor_CArg& pred_name, char *adorn_string, int adorn_length)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *temp_adorn_string = adorn_string;

    for (int i=0; i < adorn_length;i++) {
        char adorn_char = *temp_adorn_string++;
        switch (adorn_char) {
        case 'b':
            break;
        case 'f':
            break;
        case 'a': 
        {
            adorn_string[i] = 'b';
            Cor_add_disallowed_adornment(Cor_parserStruct, pred_name,
                    adorn_string, adorn_length);
            adorn_string[i] = 'f';
            Cor_add_disallowed_adornment(Cor_parserStruct, pred_name,
                    adorn_string, adorn_length);
              // Restore the string to its original state
            adorn_string[i] = 'a';
            return;
        }
        default:
            *(Cor_exEnv->err_strm)
                << "Bad character in adornment string '"
                << adorn_string << "'\n";
            return;
        }
    }

    // should get here only if all the characters in adorn_string are
    // either 'b' or 'f'

    Cor_PredAnnotations *pa = new Cor_PredAnnotations(pred_name,
        Cor_parserStruct->DisallowedAdornList);

    pa->arglist1 = pa->arglist2 = NULL;
    pa->adorn = Cor_create_adornment(adorn_string);

    Cor_parserStruct->DisallowedAdornList = pa;
}

// *************************************************************************
void Cor_add_multiset_annotation(Cor_ParserStruct *Cor_parserStruct,
    Cor_CArg *c, Cor_ArgList *al1, int check_subsum)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred;
    if (c)
      pred = *c;

    if (al1->count()) {
        FOR_EACH_ARG(arg,*al1) {
            if (! arg->is_symbol()) {
                *(Cor_exEnv->err_strm)
                    << "Error: Bad argument to multiset annotation for "
                    << pred.symbol() << endl;
            } else {
                Cor_PredAnnotations *pa = new Cor_PredAnnotations(pred,
                        Cor_parserStruct->MultisetAnnotations);

                pa->arglist1 = NULL;
                pa->check_subsum = check_subsum;
                Cor_parserStruct->MultisetAnnotations = pa;
                Cor_CArg adorn_string = *arg;

                pa->adorn = Cor_create_adornment(adorn_string);
            }

        } END_EACH_ARG

    } else {

        Cor_PredAnnotations *pa = new Cor_PredAnnotations (pred,
               Cor_parserStruct->MultisetAnnotations);

        pa->arglist1 = al1->dup();
        pa->adorn = NULL;
        pa->check_subsum = check_subsum;
        Cor_parserStruct->MultisetAnnotations = pa;
    }
}

// *************************************************************************
void Cor_add_diskrel_annotation(Cor_ParserStruct *Cor_parserStruct,
            Cor_CArg *cPredName, Cor_CArg *cAdornStr, Cor_ArgList *al1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred;
    Cor_CArg adorn_string;

    if (cPredName) 
      pred = *cPredName;

    if (cAdornStr)
      adorn_string = *cAdornStr;
 
    if (! adorn_string.is_empty()) {
        Cor_PredAnnotations *pa =
          new Cor_PredAnnotations(pred,Cor_parserStruct->diskrel_annotations);

        pa->arglist1 = al1->dup();
        Cor_parserStruct->diskrel_annotations = pa;
        pa->adorn = Cor_create_adornment(adorn_string);

    } else {
        Cor_PredAnnotations *pa =
          new Cor_PredAnnotations(pred, Cor_parserStruct->diskrel_annotations);

        pa->arglist1 = al1->dup();
        pa->adorn = NULL;
        Cor_parserStruct->diskrel_annotations = pa;
    }
}

// *************************************************************************
void Cor_add_asyncrel_annotation(Cor_ParserStruct *Cor_parserStruct,
    Cor_CArg *predName, Cor_ArgList *al1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred;

    if (predName) {
      pred = *predName;

        Cor_PredAnnotations *pa = new Cor_PredAnnotations(pred,
               Cor_parserStruct->asyncrel_annotations);

          // No need for the arglist? But save it just in case.
        pa->arglist1 = al1->dup();
        pa->adorn = NULL;
        Cor_parserStruct->asyncrel_annotations = pa;
    }
}

// *************************************************************************
void Cor_update_anno_list(Cor_CArg& name, Cor_BitVector &adorn,
        Cor_PredAnnotations *anno) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (; anno != NULL; anno = anno->next()) {
        if (anno->pred.equals(name) && anno->adorn == NULL) {
            Cor_PredAnnotations *newanno = new
                            Cor_PredAnnotations(anno->pred, NULL);
            *newanno = *anno;
            Cor_BitVector *newvec = new Cor_BitVector();
            *newvec = adorn;
            newanno->adorn = newvec;
            newanno->link(anno->next());
            anno->link(newanno);
        }
    }
}

// *************************************************************************
//
//  WARNING:  If more annotations are added, update this procedure
//  as well as the printing procedures in Unparse.C
//

void Cor_update_annotations(Cor_CArg& name, Cor_BitVector& adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_update_anno_list(name, adorn, Cor_parserStruct->MultisetAnnotations);
    Cor_update_anno_list(name, adorn, Cor_parserStruct->agg_sel_annotations);
    Cor_update_anno_list(name, adorn, Cor_parserStruct->make_index_annotations);
    Cor_update_anno_list(name, adorn, Cor_parserStruct->diskrel_annotations);
    Cor_update_anno_list(name, adorn, Cor_parserStruct->asyncrel_annotations);
}

// *************************************************************************
void Cor_annotation_error(Cor_PredAnnotations *anno, char *msg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm) << "** Error in annotation: \n";
    anno->print(Cor_exEnv->err_strm);
    *(Cor_exEnv->err_strm) << "\n       -- " << msg << ".\n";
}

// *************************************************************************
void Cor_add_index_delta_annotation(Cor_ParserStruct*, Cor_CArg*, Cor_ArgList*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm) << 
        "Sorry - add_index_delta_annotation not yet implemented\n";
}

// *************************************************************************
void Cor_add_index_delta_annotation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm) << 
        "Sorry - add_index_delta_annotation not yet implemented\n";
}

// *************************************************************************
Cor_RuleAnnoCaseDetails *Cor_make_anno_case_details(Cor_CArg *carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    carg->destroy();
    return NULL;
}

// *************************************************************************
Cor_RuleAnnoCaseDetails *Cor_make_anno_case_details(Cor_RuleAnnoCase *)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return NULL;
}

// *************************************************************************
Cor_RuleAnnoCaseName *Cor_make_anno_case_name(Cor_CArg *carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    carg->destroy();
    return NULL;
}

// *************************************************************************
Cor_RuleAnnoCase *Cor_link_anno_case_list(Cor_RuleAnnoCase*, Cor_RuleAnnoCase*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return NULL;
}

// *************************************************************************
Cor_RuleAnno *Cor_link_rule_anno(Cor_RuleAnno*, Cor_RuleAnno*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return NULL;
}

// *************************************************************************
Cor_RuleAnnoCase *Cor_make_anno_case(Cor_RuleAnnoCaseName*,
            Cor_RuleAnnoCaseDetails*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return NULL;
}

// *************************************************************************
Cor_RuleAnno *Cor_make_anno(Cor_CArg*carg, Cor_RuleAnnoCaseDetails*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    carg->destroy();
    return NULL;
}

// *************************************************************************
