// *************************************************************************
// SemiNaiveEval.C
// evaluating a rule, seperated from original RuleInfo.C -8/2/94
// *************************************************************************

#include <stdlib.h>
#include <string.h>

#include "RuleInfo.h"

#include "ArgList.h"
#include "StackMark.h"
#include "Config.h"
#include "Rule.h"
#include "Table.h"
#include "RMarkArray.h"
#include "TupleIterator.h"
#include "RelationList.h"
#include "SccInfo.h"
#include "BindEnv.h"
#include "Globals.h"
#include "Literal.h"
#include "Execute.h"
#include "CoralBreak.h"
#include "Suspend.h"
#include "Suspension.h"
 
#include "DatabaseStruct.h"
#include "Interface.h"
#include "Debug.h"

// *************************************************************************

extern Cor_DatabaseStruct *Cor_CurDB;

#ifdef DO_EXPLAIN
#ifdef DO_PROFILE
#define HANDLE_SUCCESS { \
    if ((rel->local_options & REL_EXPLAIN) && (Cor_T_Stack->count > 0))\
       Cor_do_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
        (pinfo->i_newfacts)++; \
    }
#else
#define HANDLE_SUCCESS  { \
    if ((rel->local_options & REL_EXPLAIN) && (Cor_T_Stack->count > 0)) \
       Cor_do_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
  }
#endif
#else
#ifdef DO_PROFILE
#define HANDLE_SUCCESS  { (pinfo->i_newfacts)++ ; }
#else
#define HANDLE_SUCCESS  { }
#endif
#endif

#define MARKLIST(X) ((Cor_RMark*) ((markOffsets[X]<0) ? NULL : \
            rm_array->markList[markOffsets[X]]))


#ifdef    DO_PROFILE
static long profile_rule = 0;
#endif

extern  void Cor_do_dump(Cor_Table * F, Cor_RuleInfo * rInfo,
    Cor_BindEnv * new_env, Cor_ArgList *new_arglist);
extern int Cor_backpatch_bindenvs(Cor_StackMark start,
        Cor_BindEnv *dont_rename_env,
        Cor_BindEnv *rule_env, Cor_TermLink *&renamed_vars);

extern int Cor_SimpleCopyArgs(Cor_ArgList *new_arglist,
        Cor_ArgList *old_arglist,
        Cor_BindEnv *rule_env, Cor_BindEnv *dont_rename_env,
        Cor_TermLink *&renamed_vars);

// *************************************************************************
//  Evaluates a nested loop join of the seminaive rule. The Table structure
//  F is defined in Table.h. It is a structure containing pointers to the
//  various relations that the seminaive rule will have to access. It is
//  initialized at query execution time with pointers to current relations
//  and passed to the evaluate method as an argument. The evaluate function
//  accesses all relations via the level of indirection provided by the Table.
//  rm_array is an array of RMarks that specify the current extents of the
//  'old' and 'new' relations used for seminaive evaluation. The evaluate
//  function accesses the RMarks via offsets into the rm_array specified in
//  the markOffsets array structure (class variable). Look at the code to 
//  evaluate an Scc in interp.C to understand how RMarks are manipulated.
//
//  The index 'current' counts how far we are along in the join.  It
//  represents the position of the current rhs literal being processed.
//
//  The function returns the number of new tuples generated.
// *************************************************************************

int Cor_SemiNaive::evaluate(Cor_Table *F, Cor_RMarkArray *rm_array,
            Cor_SuspendContext *s, int allow_sus )
                // if called by scc, allow_sus = 0
                // if called by exec_single_rule, allow_sus = 1
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

/*
    cout << "\nAbout to evaluate rule:\n    ";

    // SPF
    if (rInfo->rule)
        cout << *(rInfo->rule);
    else
        cout << "rule is NULL.\n";
*/

#ifdef ASYN_SOCK_REL
    if (rInfo->rule->head->predicate->asyn_sock()) {
        // skip the evaluation for async socket_rel
        return(0);
    }
#endif
    
    int index=0;
    int offset;
    int static_s;        // Indicates if s was passed in (or Null)
    Cor_Relation  *rel;
    Cor_BindEnv   *env;
    Cor_TupleIterator *tuple_its_index;
    Cor_BindEnv  **it_bindenvs = NULL;
    Cor_SuspendContext   tmp_s;        // For when no context is given
#ifdef DO_PROFILE
    Cor_ProfileInfo *pinfo;
#endif

    // for efficiency, make the following assignments
    int  rInfo_num_literals = rInfo->num_literals;
    int *rInfo_relations = rInfo->relations;

    if (!s) {

          // They are not interested in suspension?
      s = &tmp_s;
      static_s = 1;

      // initialization

      s->current = 0;
      s->answer_count = 0; // Count of number of new tuples generated.
      s->backtrack_index = 0; // used to specify next backtrack point 
   
      // Quick check for empty first relation. A relation is empty if
      // both the RMarks for the relation point to the same place.
      // The RMarks specify the extent of the relation to be scanned
      // and if they both are equal, the extent between them is empty.

      if (recursive && (rInfo_num_literals > 0)) {
        int marklist_index = 2*first;
        int tmp_mark_offset;

        if ((tmp_mark_offset = markOffsets[marklist_index]) >= 0) {
            Cor_RMark *first_mark1 = rm_array->markList[tmp_mark_offset];

            if (first_mark1) {
              if (first_mark1->rel_p) {
                if (!first_mark1->rel_p->reln->count){
                    return(0);
                }
              } else {
                // This mark is the last delta
                if (first_mark1->tindex >= first_mark1->tlist->count())
                    return(0);
              }
            }
        }
      }


#    ifdef DO_PROFILE
      // Initialize profiling information
      pinfo = &(Cor_CurrModuleInfo->Cor_GlobalProfileInfo);
      pinfo->set_rule_name(rInfo->pInfo.rule);
      profile_rule = ((Cor_exEnv->GlobalRelationOptions |
               (F->rels[rInfo_relations[rInfo_num_literals]].relPtr)->
               local_options) & REL_PROFILE_STATS);
      if (profile_rule) pinfo->init_rule();
#    endif

      if (rule_id < 0) {
        // imperative rule
        (s->rule_env) = new Cor_RuleEnv(this);
      } else {
        (s->rule_env) = F->module_env->scc_envs[this->scc->scc_id]->
                    snrule_envs[this->rule_id];
      }

      (s->tuple_its) = (s->rule_env)->tuple_its;
      (s->tuples)    = (s->rule_env)->tuples;
      it_bindenvs    = (s->rule_env)->bindenvs;
      env = & ((s->rule_env)->bindenv);

      if (rule_id < 0) {
        if (rm_array && rm_array->array_size) {
          ASSERT(unify_funct[index]);
          for (index = 0; index < rInfo_num_literals; index++) {
            (s->tuple_its)[index] =
              new Cor_TupleIterator(F->rels[rInfo->offsets[index]].relPtr,
                        rInfo->arg_list[index], env,
                        unify_funct[index], MARKLIST(2*index),
                        MARKLIST(2*index+1), &(indices[index]));
          }
        } else {
          for (index = 0; index < rInfo_num_literals; index++) {
            (s->tuple_its)[index] =
              new Cor_TupleIterator (F->rels[rInfo->offsets[index]].relPtr,
                         rInfo->arg_list[index], env,
                         unify_funct[index],
                         NULL, NULL, &(indices[index]));
          }
        }
      } else {
        if (rm_array && rm_array->array_size) {
          for (index = 0; index < rInfo_num_literals; index++) {
            (s->tuple_its)[index]->
              reset_fields(F->rels[rInfo->offsets[index]].relPtr, env,
                   MARKLIST(2*index), MARKLIST(2*index+1),
                   &(indices[index]));
          }
        } else {
          for (index = 0; index < rInfo_num_literals; index++) {
            (s->tuple_its)[index]->
              reset_fields(F->rels[rInfo->offsets[index]].relPtr, env,
                   NULL, NULL, &(indices[index]));
          }
          // NOTE:this should be moved into reset_fields!
          //F->rels[rInfo->offsets[index]].relPtr->scan_count++;
        }
      }

      s->direction = FORWARD;
    } else { 
      // some other variables derived from the above;
      env = & ((s->rule_env)->bindenv);
      it_bindenvs = (s->rule_env)->bindenvs;
      static_s = 0;
    }

    // Nested Loops Join evaluation using backtracking on failure.
    while (s->current >= 0) {

      // Exit statement from loop. When current < 0, it implies that
      // no more facts can be generated from this rule evaluation, i.e.
      // it has backtracked past its starting point.

      // Check to see if rule evaluation has been succesful.
      // This happens when a valid tuple for the last rhs literal
      // has been found. When this happens, the head fact(s) should be
      // created.

      if ( s->current >= rInfo_num_literals ) {
        index = s->current;

        // The following code allows multiple heads per rule
        for( ; rInfo_relations[index] != -1; index++) {
          offset = rInfo_relations[index];
          if (offset < 0) {
            // A negative offset implies a negated literal
            // In such a case we should delete the tuple
            // (This is a Head Delete)
            offset = (-2)-offset;
            rel = F->rels[offset].relPtr;
            Cor_ArgList del_al = rInfo->arg_list[index];
            Cor_Tuple del_tup(&del_al);
            del_tup.instantiate(env);

            //cout << "Head delete of: " << del_tup << endl;

              // Value-based delete
            delete_tuple(rel, &del_tup);
            continue;
          }
      
          // Find the relation in which the head fact should be inserted.
          rel = F->rels[offset].relPtr;
      
#       ifdef DO_PROFILE
          (pinfo->i_succesful)++;
#       endif

          // Insert the new head fact into the relation
          if (rel->insert_new(rInfo->arg_list[index], env)) {
              HANDLE_SUCCESS;
#ifdef    INSERT_VERBOSE
           cout << "INSERT: " << rel->name << "(";
           (rInfo->arg_list[index]).print(env,&cout);
           cout << ")\n";

           cout << "Arglist is: " << (rInfo->arg_list[index]) << endl;
           cout << "env is: ";
           env->printon(&cout);
           cout << endl;
#endif
              s->answer_count++;
          }
        }

        // Continue with the nested loops join
        s->current = rInfo_num_literals -1;
        s->backtrack_index = successBacktrack;            
        s->direction = BACKWARD;

        if (allow_sus && (Cor_suspension->set >= 0)) { // suspended query
            if ((Cor_suspension->set == 1)
                    || ((Cor_suspension->set == 0)
                        && (Cor_suspension->byte_count < 128))) { //suspend now

              s->answer = Cor_curUser->num_answers;

              // This needs to specify both the user and the connection
              s->conn_id = Cor_curConn->conn_id;    // who suspended it
              s->user_id = Cor_curUser->user_id;    

              s->curr_ws_name = &(Cor_CurDB->name);
              s->table = F;
              s->semi_naive = this;
              s->trail = Cor_CurTrail;       //current stack saved
              Cor_CurTrail = new Cor_Trail(200); //create new stack
           
                 // We weren't expecting to suspend earlier 
              if (static_s)
                  s = s->dup();

              Cor_suspension->suspend(s);
              return(s->answer_count);
            }
        }
        continue;
      } else {     // the entire rule has not been instantiated yet

        index = predOrders[s->current];
        tuple_its_index = (s->tuple_its)[index];

        // To do a scan of each relation on the rhs, a TupleIterator is
        // needed on that relation. If this is the first time the relation
        // is reached during the evaluation, a TupleIterator needs to be
        // created for it.

        if (s->direction) {        // FORWARD
            tuple_its_index->reset();
            s->backtrack_index = s->current;
        }
        
        if (s->current == s->backtrack_index) {
          // Get the next tuple for the current relation in
          // the nested loops join.

          (s->tuples)[index] = tuple_its_index->get_next_tuple();

          if (tuple_its_index->no_match() ^ negated[index]) {
              s->backtrack_index = btDecision[(s->direction)][s->current];
          } else {
              s->current++;
              (s->direction) = FORWARD;
              continue;
          }    
        }
        
        // else on intelligent backtracking failure or
        // on failure to find a valid tuple, backtrack to the previous
        // rhs literal.
       
        // Undo bindings BEFORE
        // deleting tuple_iterator (and the tuple_env in it).
        
        tuple_its_index->release(); // Releases state of tuple iterator
        (s->tuples)[index] = NULL;
        
        // Decrement current so that the join evaluation backtracks to the
        // previous rhs relation, and a fresh tuple for that relation will
        // be fetched in the next iteration through the for loop.
        s->current--;
        s->direction = BACKWARD;
      }
    }

    // Cleaning up at the end of the nested loops rule evaluation.

    if (rule_id < 0) {
      for (index = 0; index < rInfo_num_literals; index++) {
        delete (s->tuple_its)[index];
        (s->tuple_its)[index] = NULL;
      }
    }
    (s->stackmark0).pop_to();

#    ifdef DO_PROFILE
    if (profile_rule) {
      pinfo->exit_rule();
      _PROFILE(rInfo->pInfo, predOrders, rInfo_num_literals) ;
    }
#    endif
    
    if (rule_id < 0)
        delete (s->rule_env);

    // Set up breakpoints for rules
    if (scc) {
      int outcome =
        strcmp(Cor_brk_Info->module_name,
              (scc->module->module_info->name).symbol());

      if ((!outcome) && (Cor_brk_Info->rule_flags[scc->scc_id])) {
        *(Cor_exEnv->err_strm) << "rule number is " 
                 << rInfo->rule->rule_number << endl;
        if ((Cor_brk_Info->rule_flags[scc->scc_id])->test(
                                  rInfo->rule->rule_number+1)) {
            coral_break(Cor_Rule_String, rInfo->rule->rule_number+1);
        }
      } else if ((!outcome) && (Cor_brk_Info->all)) {
        coral_break(Cor_Rule_String, rInfo->rule->rule_number+1);
      }
    }

    int ret = s->answer_count;

    return(ret);
}

// *************************************************************************
