// *************************************************************************
//    Suspension.C
// *************************************************************************

#include "ArgList.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "Execute.h"
#include "DerivedRelation.h"
#include "BuiltinUtil.h"
#include "Globals.h"
#include "Suspend.h"
#include "Suspension.h"

#include "ClientProtocol.h"
#include "Debug.h"

// *************************************************************************

extern int Cor_currentDB(Cor_CArg& name);

// *************************************************************************
Cor_suspensionClass::Cor_suspensionClass()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

  set =-1;
  byte_count = Cor_server_page_length;
  for(int i=0; i<Cor_max_suspensions; i++) 
    context[i]=NULL;
}

// *************************************************************************
Cor_suspensionClass::~Cor_suspensionClass()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

  for(int i=0; i < Cor_max_suspensions; i++) 
    if (context[i]){
        Cor_SuspendContext *cid = context[i];
        delete cid->semi_naive;
        delete cid->table;
        delete context[i];
    }
}

// *************************************************************************
// default: list only the current user's. list(-2) for all users on this
// connection
// *************************************************************************
int Cor_suspensionClass::list(int id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int found =0;
    int cid = Cor_curConn->conn_id;

    for(int i=0; i<Cor_max_suspensions; i++) 
        if (context[i])
            if ((cid == context[i]->conn_id)
                        && ((id ==-2) || (id == context[i]->user_id))) {
                *(Cor_exEnv->out_strm)
                      << " ID=" << i
                      << " WS=" << context[i]->curr_ws_name->symbol()
                      << " SN=" << context[i]->semi_naive
                      << " TB=" << context[i]->table
                      << " ST=" << context[i]->trail
                      << " UID=" << context[i]->user_id
                      << " ANS=" << context[i]->answer << endl;
                found =1;
            }
    return(found);
}

// *************************************************************************
int Cor_suspensionClass::avail_id(void)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  for (int i=0; i<Cor_max_suspensions; i++) 
    if (context[i]==NULL) return(i);
  return(-1);            // no more id
}


// *************************************************************************
int Cor_suspensionClass::suspend(Cor_SuspendContext *s)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  int id;
  if (( id=avail_id()) != -1 ) {
    context[id]=s;
    *(Cor_exEnv->out_strm) << endl << CP_R_ID << id << "\n";
    //*(Cor_exEnv->out_strm)
            // << "Bytes Sent = " << (server_page_length - byte_count) << endl;
    set =-1;
    byte_count = Cor_server_page_length;
    return(0); //successful
  } else {
    *(Cor_exEnv->err_strm) << 
        "Maximum number of allowed suspensions has reached.\n";
    return(-1);
  }
}

// *************************************************************************
int Cor_suspensionClass::resume(int id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  Cor_StackMark stack_mark;

  if ((id>=Cor_max_suspensions)||(context[id]==NULL)) {
    *(Cor_exEnv->err_strm) << 
        "\n" << CP_CERR << "Invalid id, can't resume suspended query.\n";
    return(-1);
  }else {
#ifdef DEBUG
    *(Cor_exEnv->out_strm) << "Suspension " << id << " is resuming ";
    *(Cor_exEnv->out_strm) << "in " << context[id]->curr_ws_name->symbol()
            << "(S" << context[id]->semi_naive
            << " T" << context[id]->table
            << ") Stack=" << context[id]->trail << ":\n";
#endif

      // Check to make sure it belongs to this person
    if ((Cor_curConn && (Cor_curConn->conn_id != context[id]->conn_id)) ||
            (Cor_curUser->user_id != context[id]->user_id)) {
      *(Cor_exEnv->err_strm) << 
          "\n" << CP_CERR << "Invalid id, can't resume suspended query.\n";
      return(-1);
    }
    
    if (Cor_currentDB(*(context[id]->curr_ws_name)) == 0) 
      return -1;        // switch work space, change prompt
    strcpy(Cor_parseEnv->prompt_name, context[id]->curr_ws_name->symbol());
    strcat(Cor_parseEnv->prompt_name, coralConfig->prompt);
    *(Cor_exEnv->out_strm) << "PRMT:" << Cor_parseEnv->prompt << endl;
 
    delete Cor_CurTrail;
    Cor_CurTrail = context[id]->trail; // switch stack

    Cor_SuspendContext *cid = context[id];
    context[id]=NULL;        // free id

    if (coralConfig->server_mode) {// automatic suspension;
      set = Cor_convert_ans_2_set(cid->answer);
    }

    int num_ans = (cid->semi_naive)->evaluate(cid->table, 
                          NULL, cid,1); // continue eval
    if (num_ans == 0){
        delete cid->semi_naive;
        delete cid->table;
        delete cid;
    }

    if (Cor_exEnv->C_verbose_answer_default) 
      *(Cor_exEnv->out_strm) << "(Number of Answers = " << num_ans << ")\n";
    stack_mark.pop_to();    // clear stack for next query
    
    return(0);
  }
}

// *************************************************************************
int Cor_suspensionClass::kill(int id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  if ((id>=Cor_max_suspensions)||(context[id]==NULL)) {
    *(Cor_exEnv->err_strm) << 
        "\n" << CP_CERR << "Invalid id, can't kill suspended query.\n";
    return(-1);
  }

    // Check to make sure it belongs to this person
  if ((Cor_curConn && (Cor_curConn->conn_id != context[id]->conn_id)) ||
          (Cor_curUser->user_id != context[id]->user_id)) {
    *(Cor_exEnv->err_strm) << 
        "\n" << CP_CERR << "Invalid id, can't kill suspended query.\n";
    return(-1);
  }
                // ???? questions about allocation
  delete context[id]->trail;    // stack
  Cor_SuspendContext *cid = context[id];
  delete cid->semi_naive;
  delete cid->table;
  context[id]=NULL;        // free id
  *(Cor_exEnv->out_strm) << "Suspended query " << id << " killed.\n";
  return(0);
}

// *************************************************************************
// killall() -
// Kill all queries belonging to this user.
// *************************************************************************
int Cor_suspensionClass::killall(int conn_id, int user_id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  int i;

  for (i=0; i < Cor_max_suspensions; i++){

      // Delete suspended queries belonging to this person
    if (context[i] && (context[i]->conn_id == conn_id)
                        && (context[i]->user_id == user_id)){

      delete context[i]->trail;
      Cor_SuspendContext *cid = context[i];
      delete cid->semi_naive;
      delete cid->table;
      context[i]=NULL;
    }
  }

  return(0);
}

// *************************************************************************
// value for number of answers per query: 
//    0 - suspend according to page length.
//    n - suspend after n answers
// 
// value for set(a member in this class) before calling evaluate:
//    0 - suspend according to page length
//    n - suspend after N-1 answers. 
// *************************************************************************

int Cor_convert_ans_2_set(int ans)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  if (ans > 0) 
    return ans + 1;
  else
    return ans;
}
// *************************************************************************
