// *************************************************************************
//    QFMaterializedModuleData.C
// *************************************************************************

#include "Globals.h"
#include "QFMaterializedModuleData.h"

#include "CoralBreak.h"
#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include "Interface.h"
#include "Execute.h"
#include "HashSimpleRelation.h"
#endif

// *************************************************************************
Cor_QFMaterializedModuleData::Cor_QFMaterializedModuleData(Cor_Table *table,
    Cor_RuleInfoArray *rArray, Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    F = table;
    ruleArray = rArray;
    numSccs = 0;
    sccArray = NULL;

      // Need to set the parserStruct's pointers to NULL so
      // these items are not deleted
    multiset_annotations = Cor_parserStruct->MultisetAnnotations;
    Cor_parserStruct->MultisetAnnotations    = NULL;

    agg_sel_annotations  = Cor_parserStruct->agg_sel_annotations;
    Cor_parserStruct->agg_sel_annotations    = NULL;

    index_annotations    = Cor_parserStruct->make_index_annotations;
    Cor_parserStruct->make_index_annotations = NULL;

    diskrel_annotations  = Cor_parserStruct->diskrel_annotations;
    Cor_parserStruct->diskrel_annotations    = NULL;

    asyncrel_annotations  = Cor_parserStruct->asyncrel_annotations;
    Cor_parserStruct->asyncrel_annotations    = NULL;

    free_table_list = NULL;
    rel_options = 0;
    magic_rel = NULL;
    predicates = new Cor_ModulePred(NULL);
}

// *************************************************************************
Cor_QFMaterializedModuleData::~Cor_QFMaterializedModuleData()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (int i=0; i< numSccs; i++)
        delete sccArray[i];

    delete [] sccArray;

    delete magic_rel;

    Cor_ModulePred *curr, *prev = NULL;
    for (curr = predicates; curr; curr=curr->next) {
        delete prev;
        prev = curr;
    }
    delete prev;

    delete ruleArray;
}

// *************************************************************************
void Cor_QFMaterializedModuleData::AddScc(Cor_Scc *scc,
    Cor_Predicate *ext_magicpred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_SccInfo **temp = sccArray;

    sccArray = new Cor_SccInfo*[++numSccs];
    for (i=0; i< numSccs-1; i++)
        sccArray[i] = temp[i];

    sccArray[i] = new Cor_SccInfo(this, scc, ext_magicpred, numSccs-1);
    if (numSccs > 1) delete [] temp;
}

// *************************************************************************
void Cor_QFMaterializedModuleData::clean_cache()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i = 0; i < numSccs; sccArray[i++]->clean_cache());
}

// *************************************************************************
//  Iterator is guaranteed to be either NULL, or have a state field allocated.
//  If iterator->state->rm_Arr is NULL, it implies that this is the first
//  time this module is being invoked using this iterator

void Cor_QFMaterializedModuleData::Solve(Cor_Table *table,
    Cor_TupleIterator *iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef ASYN_SOCK_REL
    do {
#endif

    int outcome;

    outcome = strcmp(Cor_brk_Info->module_name,
        (sccArray[0]->module->module_info->name).symbol());

    if (!iterator || !(iterator->state->rm_Arr)) {
          // This is conservative.  In most cases it is not needed
        clean_cache();

        for (int i = 0; i < numSccs-1; i++) {
            if (!outcome) {
                  // Break on module
                coral_break(Cor_brk_Info->module_name,-1);
            }
            if (Cor_brk_Info->scc_flags) {
                // Break on SCC
                *(Cor_exEnv->err_strm) <<
                    "scc_id = " << sccArray[i]->scc_id << endl;
                if (!outcome && (Cor_brk_Info->scc_flags->test(
                                            sccArray[i]->scc_id+1))){
                    coral_break(Cor_Scc_String, sccArray[i]->scc_id+1);
                }
                else if ((!outcome) && (Cor_brk_Info->all)) {
                    coral_break(Cor_Scc_String, sccArray[i]->scc_id+1);
                }
            }
            sccArray[i]->Solve(table);
        }
    }

    if (Cor_brk_Info->scc_flags) {
          // Break on last scc
        *(Cor_exEnv->err_strm) <<
            "scc_id = " << sccArray[numSccs-1]->scc_id << "\n";
        if ((!outcome)&&(Cor_brk_Info->scc_flags->test(
                                sccArray[numSccs-1]->scc_id+1))){
            coral_break(Cor_Scc_String, sccArray[numSccs-1]->scc_id+1);
        }
        else if ((!outcome) && (Cor_brk_Info->all)) {
            coral_break(Cor_Scc_String, sccArray[numSccs-1]->scc_id+1);
        }
    }

    sccArray[numSccs-1]->Solve(table, iterator);

#ifdef ASYN_SOCK_REL

    if (iterator)  // Can't use Async-Sockets Rels with Lazy Evaluation
        return;

    Cor_SccInfo *scc_info = sccArray[numSccs-1];

    Cor_PredicateList *sockPredList = new Cor_PredicateList();
    for (int i = scc_info->numPreds;
                i < scc_info->numPreds+scc_info->num_lowerPreds; i++)
        if (scc_info->pred_ptrs[i]->asyn_sock()) {
            scc_info->SockRelPreds->reset();
            for (int j = 0; j < scc_info->numSockRelPreds; j++) {
                if (scc_info->pred_ptrs[i]
                            == scc_info->SockRelPreds->get_next())
                    break;
            }

            if (j == scc_info->numSockRelPreds)  {
                char magic_name[100];
                strcpy(magic_name, "m_");
                strcat(magic_name, scc_info->pred_ptrs[i]->name.symbol());

                for (int k = 0; k < numSccs; k++) {
                    int found = 0;
                    Cor_SccInfo *cur_scc = sccArray[k];
                    for (int l = 0; l <   cur_scc->numPreds
                                        + cur_scc->num_lowerPreds
                                        + cur_scc->numSockRelPreds; l++)
                        if (strcmp(cur_scc->pred_ptrs[l]->name.symbol(),
                                        magic_name) == 0) {
                            Cor_Predicate *pred = scc_info->pred_ptrs[i];
                            Cor_Relation *rel =
                              Cor_find_local_relation(pred->name, pred->arity);
                            ASSERT(rel);

                            sockPredList->append(pred);

                            pred = cur_scc->pred_ptrs[l];
                            rel = Cor_find_local_relation(pred->name,
                                pred->arity);
                            ASSERT(rel);

                            rel->send_delta(this, table, k, l, pred);
                            found = 1;
                            break;
                        }
                    if (found)
                        break;
                }
            }
        }

    for (; i <   scc_info->numPreds
               + scc_info->num_lowerPreds
               + scc_info->numSockRelPreds; i++) {
        char magic_name[100];
        strcpy(magic_name, "m_");
        strcat(magic_name, scc_info->pred_ptrs[i]->name.symbol());

        for (int k = 0; k < numSccs; k++) {
            int found = 0;
            Cor_SccInfo *cur_scc = sccArray[k];
            for (int l = 0; l <   cur_scc->numPreds
                                + cur_scc->num_lowerPreds
                                + cur_scc->numSockRelPreds; l++)
                if (!strcmp(cur_scc->pred_ptrs[l]->name.symbol(),magic_name)) {
                    Cor_Predicate *pred = scc_info->pred_ptrs[i];
                    Cor_Relation *rel =
                            Cor_find_local_relation(pred->name, pred->arity);
                    ASSERT(rel);

                    sockPredList->append(pred);

                    pred = cur_scc->pred_ptrs[l];
                    rel = Cor_find_local_relation(pred->name, pred->arity);
                    ASSERT(rel);

                    rel->send_delta(this, table, k, l, pred);
                    found = 1;
                    break;
                }

            if (found)
                break;
        }
    }

    // free those marks as what's done in SccInfo::iterateRecursiveRules()

    for (int k = 0; k < numSccs; k++) {
       Cor_RMarkArray *rm_Arr = (table->module_env->scc_envs)[k]->rm_Arr;
       sccArray[k]->slide_deltas(table,rm_Arr,0);
    }

    if (!sccArray[numSccs-1]->external_flag_on())
        break;

    // replace the following by any asyn actions
    cout << "**** EXTERNAL FLAG ON - HIT ANY KEY WHEN READY **** ";
    char ch; cin >> ch;
    //sleep(1);

    // get tuples
    sockPredList->reset();
    while (Cor_Predicate *pred = sockPredList->get_next()) {
        Cor_Relation *rel = Cor_find_local_relation(pred->name, pred->arity);
        ASSERT(rel);
        ((Cor_HashSimpleRelation*)rel)->get_tuples();
    }
    delete sockPredList;

    } while (1);

    // free the marks as what's done in SccInfo::Solve()
    for (int k = 0; k < numSccs; k++) {
        Cor_RMarkArray *rm_Arr = (table->module_env->scc_envs)[k]->rm_Arr;
        sccArray[k]->slide_deltas(table,rm_Arr,1);
    }

#endif // ASYN_SOCK_REL

}

// *************************************************************************
void Cor_QFMaterializedModuleData::print(ostream *fd, int print_preds,
                     int print_text, int print_snrules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  for (int i=0; i < numSccs; i++)
    sccArray[i]->print(fd, print_preds, print_text, print_snrules);
}

// *************************************************************************
