// *************************************************************************
//    SetSolvers.C
// *************************************************************************

#include <stdio.h>
#include <string.h>

#include "SetSolvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "HashSimpleRelation.h"
#include "BuiltinUtil.h"

#include "Debug.h"

// *************************************************************************
Cor_Tuple *Cor_CreateSetSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
            << Cor_CreateSetSymbol->symbol() << endl;
        return(NULL);
    }
    int arity;
    if ((arity = query->arity()) != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to create_set needed", NULL);
        return(NULL);
    }
    if (!(query->args())[0].is_var()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- first argument to "
            << Cor_CreateSetSymbol->symbol() << " should be free ";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return(NULL);
    }

    Cor_ArgList *alist = new Cor_ArgList(arity);
    Cor_CArg *carg = Cor_create_nested_multiset();
    (*alist)[0] = *carg;
    carg->destroy();

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return(return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_AddElemDSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
            << Cor_AddElemDSymbol->symbol() << endl;
        return(NULL);
    }
    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to add_elem_d needed", NULL);
        return(NULL);
    }
    if (!((query->args())[0]).isConstant()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- First argument to "
            << Cor_AddElemDSymbol->symbol() << " is free ";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return(NULL);
    }
    if (!(query->args())[1].is_rel()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- Second argument to "
            << Cor_AddElemDSymbol->symbol() << " is not a set: ";
        (query->args())[1].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return(NULL);
    }
    Cor_ArgList *alist = new Cor_ArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    ((*alist)[1]).rel()->insert_tuple(make_tuple(make_arglist((*alist)[0])));

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return(return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_AddElemSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
            << Cor_AddElemSymbol->symbol() << endl;
        return(NULL);
    }
    int arity;
    if ((arity = query->arity()) != 3) {
        CORAL_error(COR_BAD_ARITY, "3 arguments to add_elem needed", NULL);
        return(NULL);
    }
    if (!((query->args())[0]).isConstant()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- First argument to "
            << Cor_AddElemSymbol->symbol() << " is free ";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return(NULL);
    }
    if (!(query->args())[1].is_rel()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- Second argument to "
            << Cor_AddElemSymbol->symbol() << " is not a set: ";
        (query->args())[1].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return(NULL);
    }

    Cor_ArgList *alist = new Cor_ArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    Cor_CArg *carg = Cor_create_nested_multiset();
    ((*alist)[2]) = *carg;
    carg->destroy();

    Cor_StorageRelation *store1 = (Cor_StorageRelation *) ((*alist)[1]).rel();
    Cor_StorageRelation *store2 = (Cor_StorageRelation *) ((*alist)[2]).rel();

    store1->add_all_tuples(store2);

    ((*alist)[2]).rel()->insert_tuple(make_tuple(make_arglist((*alist)[0])));

    if (((query->args())[2]).isConstant()) {
        if (!(query->args())[2].is_rel()) {
            *(Cor_exEnv->err_strm)
                << "CORAL :: Error -- Third argument to "
                << Cor_AddElemSymbol->symbol() << " is not a set: ";
            (query->args())[2].printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            return(NULL);
        } else if (!((query->args())[2]).equals((*alist)[2]))
            return(NULL);
    }
    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return(return_tuple);
}

// *************************************************************************
Cor_BindEnv *Cor_MemberSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    Cor_Relation *rel;
    Cor_ArgList *args;
    Cor_TupleIterator *iter;

    if (iterator.ipos++ == 0) {

        iterator.stack_mark.get_mark();
        int num = iterator.arg_list.arity() - 1;

        Cor_Term term(iterator.arg_list[0], env);
        Cor_deep_dereference_term(term);

        if (term.expr.is_functor()) {
            Cor_CArg *rel_name = &(term.expr);
            rel = find_relation(rel_name->functor(),
                    rel_name->args().arity());

            if (!rel) {
                *(Cor_exEnv->err_strm) << "Relation " << rel_name->functor()
                    << " with arity " << rel_name->args().arity()
                    << " not found.\n",
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        } else if (term.expr.is_symbol()) {
            rel = find_relation(term.expr, -1);

            if (!rel) {
                *(Cor_exEnv->err_strm) << "Relation "
                    << term.expr.symbol() << " not found.\n";
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        } else if (term.expr.is_rel())
            rel = term.expr.rel();
        else {
            *(Cor_exEnv->err_strm)
                << "Wrong first argument provided to member: "
                << term.expr.symbol() << endl;
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        if ((rel->arity() > 0) && (rel->arity() != num)) {
            *(Cor_exEnv->err_strm) << "Relation arity does not match.\n";
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        args = new Cor_ArgList(num);
        for (i = 1; i <= num; i++)
            (*args)[i - 1] = iterator.arg_list[i];

        iter = new Cor_TupleIterator(rel, *args, env);
        iter->del_arglist = args;    // The iterator will delete this later.
        iterator.answer_iterator = iter;

    } else {
        iterator.stack_mark.pop_to();
        iter = iterator.answer_iterator;
        rel = iter->relation;
    }

    Cor_BindEnv *result = iter->get_next();
    if (iter->no_match()) {
        delete iter;
        iterator.set_no_match();
        return(NULL);
    } else {
        iterator.reset_no_match();
        return(result);
    }
}

// *************************************************************************
Cor_BindEnv *Cor_MemTupleSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *env = iterator.bindenv;

    Cor_Relation *rel;
    Cor_TupleIterator *iter;

    if (iterator.ipos++ == 0) {
        iterator.stack_mark.get_mark();
        //int num = iterator.arg_list.arity() - 1;

        Cor_Term term(iterator.arg_list[0], env);
        Cor_deep_dereference_term(term);
        if (term.expr.is_functor()) {
            Cor_CArg *rel_name = &(term.expr);
            rel = find_relation(rel_name->functor(),
                    rel_name->arity());

            if (!rel) {
                *(Cor_exEnv->err_strm)
                    << "Relation " << rel_name->functor()
                    << " with arity " << rel_name->arity()
                    << " not found.\n";
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        } else if (term.expr.is_symbol()) {
            rel = find_relation(term.expr, -1);

            if (!rel) {
                *(Cor_exEnv->err_strm)
                    << "Relation " << term.expr.symbol()
                    << " not found.\n";
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        } else if (term.expr.is_rel())
            rel = term.expr.rel();
        else {
            *(Cor_exEnv->err_strm) << 
                         "Wrong first argument provided to MemTuple.\n";
            iterator.set_no_match();
            return(iterator.bindenv);
        }
        Cor_ArgList *args = make_vararglist(rel->arity());
        Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
        iter = new Cor_TupleIterator(rel, *args, env0);
        iter->stack_mark.get_mark();
        iterator.answer_iterator = iter;
    } else {
        iterator.stack_mark.pop_to();
        iter = iterator.answer_iterator;
        rel = iter->relation;
    }

    for (;;) {
        Cor_Tuple *ret_tuple = iter->get_next_tuple();
        if (iter->no_match()) {
            iterator.set_no_match();
            return(NULL);
        }
        Cor_CArg *func = make_arg("", ret_tuple->_args);

        Cor_Term lterm(iterator.arg_list[1], env);
        Cor_Term rterm(func, iter->bindenv);
        if (Cor_unify_args(lterm, rterm)) {
            iterator.reset_no_match();
            return(iterator.bindenv);
        }
    }
}

// *************************************************************************
Cor_BindEnv *Cor_MultiUnionSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        *(Cor_exEnv->err_strm) << 
                        "Error: unionmax: wrong number of arguments.\n";
        return(NULL);
    }

    Cor_Term terms[3];
    for (i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "MultiUnionSolver");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "MultiUnionSolver");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "MultiUnionSolver");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[1]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[1]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[1]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[1]->arity());
    Cor_BindEnv *env3 = new Cor_BindEnv(rels[1]->arity());
    Cor_ArgList *args3 = make_vararglist(rels[1]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                      "Error: Union: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Relation *union_rel = new Cor_HashSimpleRelation(rels[0]->arity());

    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[1], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);
    Cor_TupleIterator *iter3 = new Cor_TupleIterator(union_rel, *args3, env3);

    Cor_Tuple *ret_tuple;
    for (;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        union_rel->insert_new(ret_tuple);
    }

    // second relation
    Cor_Tuple *ret_tuple2;
    int _arity = rels[1]->arity();
    for (;;) {
        ret_tuple = iter1->get_next_tuple();
        if (iter1->no_match())
            break;

        // if there is a match find out how many copies there are
        Cor_StackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple->arg(i);
        int rel2_count = 0;
        //    looping through the second relation
        for (;;) {
            ret_tuple2 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        // find out how many copies are there in the union
        Cor_StackMark stack_mark3;
        iter3->reset();
        for (i = 0; i < _arity; i++)
            (iter3->arg_list)[i] = ret_tuple->arg(i);
        int union_count = 0;
        // loop through the union
        for (;;) {
            ret_tuple2 = iter3->get_next_tuple();
            if (iter3->no_match()) {
                stack_mark3.pop_to();
                iter3->release();
                break;
            }
            union_count++;
        }

        // insert number of extra copies in union
        for (i = 0; i < (rel2_count - union_count); i++)
            union_rel->insert_new(ret_tuple);
    }


    stack_mark0.pop_to();
    Cor_Term result_term;
    result_term.expr = union_rel;
    result_term.bindenv = NULL;
    if (!Cor_unify_args(result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_UnionSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        *(Cor_exEnv->err_strm) << "Error: unionsum: wrong number arguments.\n";
        return(NULL);
    }

    Cor_Term terms[3];
    for (i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);

    Cor_BindEnv *env1 = new Cor_BindEnv(rels[1]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[1]->arity());
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[1], *args1, env1);

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                      "Error: Union: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Relation *union_rel = new Cor_HashSimpleRelation(rels[0]->arity());

    Cor_Tuple *ret_tuple;
    for (;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
            break;
        union_rel->insert_new(ret_tuple);
    }

    for (;;) {
        ret_tuple = iter1->get_next_tuple();
        if (iter1->no_match())
            break;
        union_rel->insert_new(ret_tuple);
    }

    stack_mark0.pop_to();
    Cor_Term result_term;
    result_term.expr = union_rel;
    result_term.bindenv = NULL;
    if (!Cor_unify_args(result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_IntersectionSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        *(Cor_exEnv->err_strm) << "Error: Inter: wrong number arguments.\n";
        return(NULL);
    }

    Cor_Term terms[3];
    for (i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env3 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args3 = make_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
               "Error: Intersection: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Relation *intersection_rel = new
    Cor_HashSimpleRelation(rels[0]->arity());

    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[0], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);
    Cor_TupleIterator *iter3 = new Cor_TupleIterator(intersection_rel,
                                                     *args3, env3);
    int _arity = rels[1]->arity();
    // loop through outer relation
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // see if tuple already in intersection
        Cor_StackMark stack_mark3;
        iter3->reset();
        for (i = 0; i < _arity; i++)
            (iter3->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter3->get_next_tuple();
        if (!iter3->no_match()) {
            stack_mark3.pop_to();
            iter3->release();
            continue;
        }
        // loop through relation again to count number of copies
        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        // loop through second relation to find number of copies
        Cor_StackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for (;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        if (rel2_count == 0)
            continue;

        int min;
        if (rel2_count < rel1_count)
            min = rel2_count;
        else
            min = rel1_count;

        for (i = 0; i < min; i++)
            intersection_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    Cor_Term result_term;
    result_term.expr = intersection_rel;
    result_term.bindenv = NULL;
    if (!Cor_unify_args(result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_DiffSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        *(Cor_exEnv->err_strm) << 
                         "Error: Difference: wrong number arguments.\n";
        return(NULL);
    }

    Cor_Term terms[3];
    for (i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env3 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args3 = make_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
               "Error: Intersection: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Relation *diff_rel = new Cor_HashSimpleRelation(rels[0]->arity());

    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[0], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);
    Cor_TupleIterator *iter3 = new Cor_TupleIterator(diff_rel, *args3, env3);
    int _arity = rels[1]->arity();

    // loop through outer relation
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        //    see if tuple already in difference
        Cor_StackMark stack_mark3;
        iter3->reset();
        for (i = 0; i < _arity; i++)
            (iter3->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter3->get_next_tuple();
        if (!iter3->no_match()) {
            stack_mark3.pop_to();
            iter3->release();
            continue;
        }
        // loop through relation again to count number of copies
        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        // loop through second relation to find number of copies
        Cor_StackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for (;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        for (i = 0; i < (rel1_count - rel2_count); i++)
            diff_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    Cor_Term result_term;
    result_term.expr = diff_rel;
    result_term.bindenv = NULL;
    if (!Cor_unify_args(result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_SubsetSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << 
                          "Error: subset: wrong number of arguments.\n";
        return(NULL);
    }

    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                     "Error: Subset: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Relation *subset_rel = new Cor_HashSimpleRelation(rels[0]->arity());
    if (subset_rel)                // Eliminate "unused" warning
        ;                        // Nothing


    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;
    int no_match = 0;
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[0], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);
    int _arity = rels[1]->arity();
    //    loop through outer relation
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through relation again to count number of copies
        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        /* loop through second relation to find number of copies */
        Cor_StackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for (;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }
        if (rel1_count > rel2_count) {
            no_match = 1;
            break;
        }
    }

    stack_mark0.pop_to();
    if (!no_match)
        iterator.reset_no_match();
    else {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
    }
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_MakeSetSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << "Error: make_set: wrong number arguments.\n";
        return(NULL);
    }
    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rel;
    switch (terms[0].Kind()) {

    case REL_ARG:
        rel = terms[0].expr.rel();
        break;

    case FUNCT_ARG:
        rel = find_relation(terms[0].expr.functor(),
                            terms[0].expr.arity());
        if (rel == NULL) {
            CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
        break;

    case SYMBOL_ARG:
        rel = find_relation(terms[0].expr, -1);
        if (rel == NULL) {
            CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
        break;

    default:
        CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args0 = make_vararglist(rel->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args1 = make_vararglist(rel->arity());

    Cor_Relation *set_rel = new Cor_HashSimpleRelation(rel->arity());

    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rel, *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(set_rel, *args1, env1);

    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;
    int _arity = rel->arity();
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter1->get_next_tuple();
        if (!iter1->no_match()) {
            stack_mark1.pop_to();
            iter1->release();
            continue;
        }
        stack_mark1.pop_to();
        iter1->release();
        set_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    Cor_Term result_term;
    result_term.expr = set_rel;
    result_term.bindenv = NULL;
    if (!Cor_unify_args(result_term, terms[1]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_UnionDSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << 
                      "Error: unionsum_d: wrong number of arguments.\n";
        return(NULL);
    }

    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                    "Error: union_d: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Tuple *ret_tuple;
    for (;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        rels[1]->insert_new(ret_tuple);
    }


    stack_mark0.pop_to();

    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_MultiUnionDSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << 
                          "Error: unionmax_d wrong number arguments.\n";
        return(NULL);
    }

    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                 "Error: unionmax_d: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[0]->arity());

    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[0], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);

    int _arity = rels[1]->arity();

    // loop through first relation
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through both relations to count number of copies */

        // first relation...

        Cor_StackMark stack_mark1;
        iter1->reset();

        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        int rel1_count = 0;

        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        // second relation...


        Cor_StackMark stack_mark2;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

        int rel2_count = 0;

        for (;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        int diff = rel1_count - rel2_count;
        if (diff <= 0)
            continue;                // 2nd rel already has correct number

        // else insert diff many copies of ret_tuple0 into 2nd rel...

        while (diff-- > 0) {
            rels[1]->insert_new(ret_tuple0);
        }
    }

    stack_mark0.pop_to();
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_IntersectionDSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << "Error: inter_d wrong number arguments.\n";
        return(NULL);
    }

    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
                    "Error: inter_d: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rels[0]->arity());
    Cor_BindEnv *env2 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[0]->arity());
    Cor_ArgList *args2 = make_vararglist(rels[0]->arity());

    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[0], *args1, env1);
    Cor_TupleIterator *iter2 = new Cor_TupleIterator(rels[1], *args2, env2);

    int _arity = rels[1]->arity();

    // loop through outer relation
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through both relations to count number of copies

        // first relation...

        Cor_StackMark stack_mark1;
        iter1->reset();

        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        int rel1_count = 0;

        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        // second relation...


        Cor_StackMark stack_mark2;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

        int rel2_count = 0;

        for (;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        int diff = rel2_count - rel1_count;
        if (diff <= 0)
            continue;                // 2nd rel already has correct number

        // else delete diff many copies of ret_tuple0 from 2nd rel...

        Cor_StackMark stack_mark3;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

        while (diff-- > 0) {
            ret_tuple1 = iter2->get_next_tuple();
            delete_id_tuple(rels[1],ret_tuple1);
        }

        stack_mark3.pop_to();
        iter2->release();
    }

    stack_mark0.pop_to();
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_DiffDSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        *(Cor_exEnv->err_strm) << 
                    "Error: difference_d: wrong number of arguments.\n";
        return(NULL);
    }

    Cor_Term terms[2];
    for (i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        Cor_deep_dereference_term(terms[i]);
    }

    Cor_Relation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].Kind()) {

        case REL_ARG:
            rels[i] = terms[i].expr.rel();
            break;

        case FUNCT_ARG:
            rels[i] = find_relation(terms[i].expr.functor(),
                               terms[i].expr.arity());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        case SYMBOL_ARG:
            rels[i] = find_relation(terms[i].expr, -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rels[0]->arity());
    Cor_ArgList *args0 = make_vararglist(rels[0]->arity());
    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rels[0], *args0, env0);

    Cor_BindEnv *env1 = new Cor_BindEnv(rels[1]->arity());
    Cor_ArgList *args1 = make_vararglist(rels[1]->arity());
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rels[1], *args1, env1);

    if (rels[0]->arity() != rels[1]->arity()) {
        *(Cor_exEnv->err_strm) << 
               "Error: difference_d: Both sets must have equal arity\n";
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;
    int _arity = rels[0]->arity();

    // scan all of rels[0]
    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // delete 1 copy of ret_tuple0 from rels[1] if it exists.

        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter1->get_next_tuple();
        if (!(iter1->no_match())) {
            delete_id_tuple(rels[1],ret_tuple1);
        }

        stack_mark1.pop_to();
        iter1->release();
    }

    stack_mark0.pop_to();
    return(env);
}

// *************************************************************************
Cor_BindEnv *Cor_MakeSetDSolver(Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 1) {
        *(Cor_exEnv->err_strm) << 
             "Error: make_set_d: wrong number of arguments provided.\n";
        return(NULL);
    }
    Cor_Term terms;
    terms.expr = iterator.arg_list[0];
    terms.bindenv = env;
    Cor_deep_dereference_term(terms);

    Cor_Relation *rel;
    switch (terms.Kind()) {

    case REL_ARG:
        rel = terms.expr.rel();
        break;

    case FUNCT_ARG:
        rel = find_relation(terms.expr.functor(),
                            terms.expr.arity());
        if (rel == NULL) {
            CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
        break;

    case SYMBOL_ARG:
        rel = find_relation(terms.expr, -1);
        if (rel == NULL) {
            CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
            iterator.stack_mark.pop_to();
            return(NULL);
        }
        break;

    default:
        CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
        iterator.stack_mark.pop_to();
        return(NULL);
    }
    Cor_StackMark stack_mark0;

    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args0 = make_vararglist(rel->arity());
    Cor_BindEnv *env1 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args1 = make_vararglist(rel->arity());

    Cor_TupleIterator *iter0 = new Cor_TupleIterator(rel, *args0, env0);
    Cor_TupleIterator *iter1 = new Cor_TupleIterator(rel, *args1, env1);

    Cor_Tuple *ret_tuple0;
    Cor_Tuple *ret_tuple1;

    int _arity = rel->arity();

    for (;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        Cor_StackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        // want to delete all but 1 copy of ret_tuple0. know there is at least
        // one, so set a new iterator to point at the first one, then delete
        // the rest.

        ret_tuple1 = iter1->get_next_tuple();
        // this is the first one.

        // now delete the rest...
        for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            // delete the tuple...
            delete_id_tuple(rel,ret_tuple1);
        }
    }

    stack_mark0.pop_to();

    return(env);
}

// *************************************************************************
